#+xcvb (module (:build-depends-on () :depends-on ("/asdf")))
#| ;;; cl-launch 3.22 lisp header

|# ;;;; Silence our lisp implementation for quiet batch use...

#| We'd like to evaluate as little as possible of the code without compilation.
 This poses a typical bootstrapping problem: the more sophistication we want
 to distinguish what to put where in what dynamic environment, the more code
 we have to evaluate before we may actually load compiled files. And, then,
 it is a waste of time to try to compile said code into a file. Moving things
 to the shell can only help so much, and reduces flexibility. Our best bet is
 to tell sbcl or cmucl to not try to optimize too hard.
|#
(eval-when (:compile-toplevel :load-toplevel :execute)
(setf *print-readably* nil ; allegro 5.0 notably will bork without this
      *print-level* nil
      *load-verbose* nil *compile-verbose* nil *compile-print* nil *load-print* nil)
(defvar cl-user::*asdf-directory-pathname*
  (merge-pathnames #p"cl/asdf/" (user-homedir-pathname))
  "directory where ASDF is installed, if not provided by your implementation.")

(unless (member :asdf *features*)
  (ignore-errors (funcall 'require "asdf")))
(unless (member :asdf *features*)
  (ignore-errors (load (merge-pathnames "build/asdf.lisp" cl-user::*asdf-directory-pathname*))))
(unless (member :asdf *features*)
  (error "Could not load ASDF."))

(in-package :asdf))
;; Make sure we use the latest ASDF available.
(eval-when (:compile-toplevel :load-toplevel :execute)

(ignore-errors
 (pushnew cl-user::*asdf-directory-pathname* *central-registry*)
 (handler-bind ((warning #'muffle-warning))
   (operate 'load-op :asdf :verbose nil)))

(unless (member :asdf3 *features*)
  (unless (asdf::version-satisfies (asdf::asdf-version) "2.15")
    (error "CL-Launch requires ASDF 2.015 or later")) ; fallback feature
  (asdf:load-system :asdf-driver))

;;;; Ensure package hygiene
#+gcl2.6
(unless (find-package :cl-launch) (make-package :cl-launch :use '(:lisp)))
#-gcl2.6
(defpackage :cl-launch
  (:use :common-lisp :asdf/driver :asdf)
  (:export #:compile-and-load-file))

(in-package :cl-launch))
;;;; CL-Launch Initialization code
(progn
(defvar *cl-launch-file* nil) ;; name of this very file
(defvar *verbose* nil)
(progn
  (defun call-with-new-file (n f)
    (with-open-file (o n :direction :output :if-exists :error :if-does-not-exist :create)
      (funcall f o)))
  (defun dump-stream-to-file (i n)
    (call-with-new-file n (lambda (o) (copy-stream-to-stream i o))))
  (defun dump-sexp-to-file (x n)
    (call-with-new-file
     n
     (lambda (o) (write x :stream o :pretty t :readably t))))
  (defvar *temporary-filenames* nil)
  (defvar *temporary-file-prefix*
    (format nil "~Acl-launch-~A-" asdf/driver:*temporary-directory* (getenvp "CL_LAUNCH_PID")))
  (defun make-temporary-filename (x)
    (concatenate 'string *temporary-file-prefix* x))
  (defun register-temporary-filename (n)
    (push n *temporary-filenames*)
    n)
  (defun temporary-filename (x)
    (register-temporary-filename (make-temporary-filename x)))
  (defun temporary-file-from-foo (dumper arg x)
    (let ((n (temporary-filename x)))
      (funcall dumper arg n)
      n))
  (defun temporary-file-from-stream (i x)
    (temporary-file-from-foo #'dump-stream-to-file i x))
  (defun temporary-file-from-string (i x)
    (temporary-file-from-foo #'princ i x))
  (defun temporary-file-from-sexp (i x)
    (temporary-file-from-foo #'dump-sexp-to-file i x))
  (defun temporary-file-from-file (f x)
    (with-open-file (i f :direction :input :if-does-not-exist :error)
      (temporary-file-from-stream i x)))
  (defun ensure-lisp-file-name (x &optional (name "load.lisp"))
    (let ((p (pathname x)))
      (if (equal (pathname-type p) "lisp")
          p
          (temporary-file-from-file p name))))
  (defun ensure-lisp-file (x &optional (name "load.lisp"))
    (cond
      ((eq x t) (temporary-file-from-stream *standard-input* "load.lisp"))
      ((streamp x) (temporary-file-from-stream x "load.lisp"))
      ((eq x :self) (ensure-lisp-file-name *cl-launch-file* name))
      (t (ensure-lisp-file-name x name))))
  (defun cleanup-temporary-files ()
    (loop :for n = (pop *temporary-filenames*)
          :while n :do
          (ignore-errors (delete-file n)))))
(defun file-newer-p (new-file old-file)
  "Returns true if NEW-FILE is strictly newer than OLD-FILE."
  (> (file-write-date new-file) (file-write-date old-file)))
(defun compile-and-load-file (source &key force-recompile
                              (verbose *verbose*) (load t)
                              output-file)
  "compiles and load specified SOURCE file, if either required by keyword
argument FORCE-RECOMPILE, or not yet existing, or not up-to-date.
Keyword argument VERBOSE specifies whether to be verbose.
Returns two values: the fasl path, and T if the file was (re)compiled"

  ;; When in doubt, don't trust - recompile. Indeed, there are
  ;; edge cases cases when on the first time of compiling a simple
  ;; auto-generated file (e.g. from the automated test suite), the
  ;; fasl ends up being written to disk within the same second as the
  ;; source was produced, which cannot be distinguished from the
  ;; reverse case where the source code was produced in the same split
  ;; second as the previous version was done compiling. Could be
  ;; tricky if a big system needs be recompiled as a dependency on an
  ;; automatically generated file, but for cl-launch those
  ;; dependencies are not detected anyway (BAD). If/when they are, and
  ;; lacking better timestamps than the filesystem provides, you
  ;; should sleep after you generate your source code.
  #+(and gcl (not gcl2.6))
  (setf source (ensure-lisp-file-name source (concatenate 'string (pathname-name source) ".lisp")))
  (let* ((truesource (truename source))
         (fasl (or output-file (compile-file-pathname* truesource)))
         (compiled-p
          (when (or force-recompile
                    (not (probe-file fasl))
                    (not (file-newer-p fasl source)))
            (ensure-directories-exist fasl)
            (multiple-value-bind (path warnings failures)
                (compile-file* truesource :output-file fasl)
              (declare (ignorable warnings failures))
              (unless (equal (truename fasl) (truename path))
                (error "CL-Launch: file compiled to ~A, expected ~A" path fasl))
              (when failures
                (error "CL-Launch: failures while compiling ~A" source)))
            t)))
    (when load
      (load* fasl :verbose verbose))
    (values fasl compiled-p)))
(defun load-file (source &key output-file)
  (declare (ignorable output-file))
  #-(or gcl2.6 (and ecl (not dlopen)))
  (compile-and-load-file source :verbose *verbose* :output-file output-file)
  #+gcl2.6
  (let* ((pn (parse-namestring source))) ; when compiling, gcl 2.6 will always
    (if (pathname-type pn) ; add a type .lsp if type is missing, to avoid compilation
      (compile-and-load-file source :verbose *verbose* :output-file output-file)
      (load source :verbose *verbose*)))
  #+(and ecl (not dlopen))
  (load source :verbose *verbose*))

(defun compute-arguments ()
  (setf *cl-launch-file* (getenvp "CL_LAUNCH_FILE")
        *verbose* (when (getenvp "CL_LAUNCH_VERBOSE") t)))

;; We provide cl-launch, no need to go looking for it further!
(unless (fboundp 'asdf::register-preloaded-system)
  (eval
   '(progn
      (defvar asdf::*preloaded-systems* (make-hash-table :test 'equal))
      (defun asdf::sysdef-preloaded-system-search (requested)
        (let ((name (coerce-name requested)))
          (multiple-value-bind (keys foundp) (gethash name asdf::*preloaded-systems*)
            (when foundp
              (apply 'make-instance 'system :name name :source-file (getf keys :source-file) keys)))))
      (defun asdf::register-preloaded-system (system-name &rest keys)
        (setf (gethash (coerce-name system-name) asdf::*preloaded-systems*) keys))
      (asdf::appendf asdf:*system-definition-search-functions* '(asdf::sysdef-preloaded-system-search)))))

(asdf::register-preloaded-system "cl-launch")

(defun do-build-and-load (load system restart final init quit)
  (etypecase load
    (null nil)
    ((eql t) (load* nil))
    (stream (load* load))
    ((eql :self) (load-file *cl-launch-file*))
    ((or pathname string) (load-file load)))
  (when system
    (load-systems system))
  (when final
    (load-from-string final))
  (setf *image-prelude* init
        *image-entry-point* (when restart (ensure-function restart))
        *lisp-interaction* (not quit)))

(defun build-and-load (load system restart final init quit)
  (unwind-protect
       (do-build-and-load load system restart final init quit)
    #+(and gcl (not gcl2.6))
    (cleanup-temporary-files)))


(defun build-and-run (load system restart final init quit)
  (build-and-load load system restart final init quit)
  (restore-image))

#-ecl
(defun build-and-dump (dump load system restart final init quit)
  (build-and-load load system restart final init quit)
  (dump-image dump :executable (getenvp "CL_LAUNCH_STANDALONE"))
  (quit 0))

;;; Attempt at compiling directly with ECL's make-build and temporary wrapper asd's
#+ecl (defvar *in-compile* nil)
#+ecl
(defun build-and-dump (dump load system restart final init quit)
  (unwind-protect
       (let* ((*compile-verbose* *verbose*)
              (*in-compile* t)
              (c::*suppress-compiler-warnings* (not *verbose*))
              (c::*suppress-compiler-notes* (not *verbose*))
              (*features* (remove :cl-launch *features*))
              (header (or *compile-file-pathname* *load-pathname* (getenvp "CL_LAUNCH_HEADER")))
              (header-file (ensure-lisp-file header "header.lisp"))
              (load-file (when load (ensure-lisp-file load "load.lisp")))
              (standalone (getenvp "CL_LAUNCH_STANDALONE"))
              (init-code
               `(progn
                  (unless *in-compile*
                    (setf
                     *package* (find-package :cl-user)
                     *load-verbose* nil
                     *dumped* ,(if standalone :standalone :wrapped)
                     *arguments* nil
                     ;;,(symbol* :asdf :*source-registry*) nil
                     ;;,(symbol* :asdf :*output-translations*) nil
                     ,@(when restart `(*restart* (read-function ,restart)))
                     *init-forms* ,init)
                    ,@(unless quit `(*quit* nil)))
                 ,(if standalone '(asdf/image:restore-image) '(si::top-level))))
              (final-file (temporary-file-from-string final "final.lisp"))
              (init-file (temporary-file-from-sexp init-code "init.lisp"))
              (prefix-sys (pathname-name (temporary-filename "prefix")))
              (program-sys (pathname-name (temporary-filename "program")))
              (prefix-sysdef
               `(,(symbol* :asdf :defsystem) ,prefix-sys
                 :depends-on () :serial t
                 :components ((:file "header" :pathname ,(truename header-file))
                              ,@(when load-file `((:file "load" :pathname ,(truename load-file)))))))
              (program-sysdef
               `(,(symbol* :asdf :defsystem) ,program-sys
                 :serial t
                 :depends-on (,prefix-sys
                              ,@(when system `(,system))
                              ,prefix-sys) ;; have the prefix first, whichever order asdf traverses
                 :components ((:file "final" :pathname ,(truename final-file))
                              (:file "init" :pathname ,(truename init-file)))))
              (prefix-asd (temporary-file-from-sexp prefix-sysdef "prefix.asd"))
              (program-asd (temporary-file-from-sexp program-sysdef "program.asd")))
         (load prefix-asd)
         (load program-asd)
         (call :asdf :make-build program-sys :type :program)
         (si:system (format nil "cp -p ~S ~S"
                            (namestring (first (call :asdf :output-files
                                                     (make-instance (symbol* :asdf :program-op))
                                                     (call :asdf :find-system program-sys))))
                            dump)))
    (cleanup-temporary-files))
  (quit))

;;(handler-bind (#+ecl (si::simple-package-error (lambda (x) (declare (ignore x)) (invoke-restart 'continue)))) (format *trace-output* "Enabling some debugging~%") #+ecl (trace c::builder c::build-fasl c:build-static-library c:build-program ensure-lisp-file-name ensure-lisp-file cleanup-temporary-files delete-package) #+ecl (setf c::*compiler-break-enable* t) (trace load-file load-systems build-and-dump build-and-run run compile-and-load-file load compile-file) (setf *verbose* t *load-verbose* t *compile-verbose* t))

(defun run (&key source-registry load system dump restart final init (quit 0))
  (pushnew :cl-launched *features*)
  (compute-arguments)
  (when source-registry (initialize-source-registry source-registry))
  (if dump
      (build-and-dump dump load system restart final init quit)
      (build-and-run load system restart final init quit)))

(pushnew :cl-launch *features*))

(cl-launch::compute-arguments)

;;; END OF CL-LAUNCH LISP HEADER
